@@ -4,32 +4,26 @@
  *	This file contains routines for converting file names betwen
  *	native and network form.
  *
- * Copyright (c) 1995-1996 Sun Microsystems, Inc.
+ * Copyright (c) 1995-1998 Sun Microsystems, Inc.
+ * Copyright (c) 1998-1999 by Scriptics Corporation.
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclFileName.c,v 1.4 1999/02/03 19:12:26 stanton Exp $
+ * RCS: @(#) $Id: tclFileName.c,v 1.6 1999/04/16 00:46:46 stanton Exp $
  */
 
 #include "tclInt.h"
 #include "tclPort.h"
 #include "tclRegexp.h"
 
-/*
- * This variable indicates whether the cleanup procedure has been
- * registered for this file yet.
- */
-
-static int initialized = 0;
-
 /*
  * The following regular expression matches the root portion of a Windows
  * absolute or volume relative path.  It will match both UNC and drive relative
  * paths.
  */
 
-#define WIN_ROOT_PATTERN "^(([a-zA-Z]:)|[/\\][/\\]+([^/\\]+)[/\\]+([^/\\]+)|([/\\]))([/\\])*"
+#define WIN_ROOT_PATTERN "^(([a-zA-Z]:)|[/\\\\][/\\\\]+([^/\\\\]+)[/\\\\]+([^/\\\\]+)|([/\\\\]))([/\\\\])*"
 
 /*
  * The following regular expression matches the root portion of a Macintosh
@@ -44,8 +38,13 @@ static int initialized = 0;
  * for use in filename matching.
  */
 
-static regexp *winRootPatternPtr = NULL;
-static regexp *macRootPatternPtr = NULL;
+typedef struct ThreadSpecificData {
+    int initialized;
+    Tcl_Obj *winRootPatternPtr;
+    Tcl_Obj *macRootPatternPtr;
+} ThreadSpecificData;
+
+static Tcl_ThreadDataKey dataKey;
 
 /*
  * The following variable is set in the TclPlatformInit call to one
@@ -54,6 +53,16 @@ static regexp *macRootPatternPtr = NULL;
 
 TclPlatformType tclPlatform = TCL_PLATFORM_UNIX;
 
+/*
+ * The "globParameters" argument of the globbing functions is an 
+ * or'ed combination of the following values:
+ */
+
+#define GLOBMODE_NO_COMPLAIN      1
+#define GLOBMODE_JOIN             2
+#define GLOBMODE_DIR              4
+#define GLOBMODE_DIR_ONLY         8
+
 /*
  * Prototypes for local procedures defined in this file:
  */
@@ -63,6 +72,7 @@ static char *		DoTildeSubst _ANSI_ARGS_((Tcl_Interp *interp,
 static CONST char *	ExtractWinRoot _ANSI_ARGS_((CONST char *path,
 			    Tcl_DString *resultPtr, int offset));
 static void		FileNameCleanup _ANSI_ARGS_((ClientData clientData));
+static void		FileNameInit _ANSI_ARGS_((void));
 static int		SkipToChar _ANSI_ARGS_((char **stringPtr,
 			    char *match));
 static char *		SplitMacPath _ANSI_ARGS_((CONST char *path,
@@ -72,6 +82,34 @@ static char *		SplitWinPath _ANSI_ARGS_((CONST char *path,
 static char *		SplitUnixPath _ANSI_ARGS_((CONST char *path,
 			    Tcl_DString *bufPtr));
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * FileNameInit --
+ *
+ *	This procedure initializes the patterns used by this module.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	Compiles the regular expressions.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+FileNameInit()
+{
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
+    if (!tsdPtr->initialized) {
+	tsdPtr->initialized = 1;
+	tsdPtr->winRootPatternPtr = Tcl_NewStringObj(WIN_ROOT_PATTERN, -1);
+	tsdPtr->macRootPatternPtr = Tcl_NewStringObj(MAC_ROOT_PATTERN, -1);
+	Tcl_CreateThreadExitHandler(FileNameCleanup, NULL);
+    }
+}
+
 /*
  *----------------------------------------------------------------------
  *
@@ -93,15 +131,10 @@ static void
 FileNameCleanup(clientData)
     ClientData clientData;	/* Not used. */
 {
-    if (winRootPatternPtr != NULL) {
-	ckfree((char *)winRootPatternPtr);
-        winRootPatternPtr = (regexp *) NULL;
-    }
-    if (macRootPatternPtr != NULL) {
-	ckfree((char *)macRootPatternPtr);
-        macRootPatternPtr = (regexp *) NULL;
-    }
-    initialized = 0;
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
+    Tcl_DecrRefCount(tsdPtr->winRootPatternPtr);
+    Tcl_DecrRefCount(tsdPtr->macRootPatternPtr);
+    tsdPtr->initialized = 0;
 }
 
 /*
@@ -132,47 +165,51 @@ ExtractWinRoot(path, resultPtr, offset)
 				 * stored. */
 {
     int length;
+    Tcl_RegExp re;
+    char *dummy, *tail, *drive, *hostStart, *hostEnd, *shareStart,
+	*shareEnd, *lastSlash;
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
     /*
      * Initialize the path name parser for Windows path names.
      */
 
-    if (winRootPatternPtr == NULL) {
-	winRootPatternPtr = TclRegComp(WIN_ROOT_PATTERN);
-	if (!initialized) {
-	    Tcl_CreateExitHandler(FileNameCleanup, NULL);
-	    initialized = 1;
-	}
-    }
+    FileNameInit();
+
+    re = Tcl_GetRegExpFromObj(NULL, tsdPtr->winRootPatternPtr, REG_ADVANCED);
 
     /*
      * Match the root portion of a Windows path name.
      */
 
-    if (!TclRegExec(winRootPatternPtr, (char *) path, (char *) path)) {
+    if (!Tcl_RegExpExec(NULL, re, path, path)) {
 	return path;
     }
 
     Tcl_DStringSetLength(resultPtr, offset);
 
-    if (winRootPatternPtr->startp[2] != NULL) {
-	Tcl_DStringAppend(resultPtr, winRootPatternPtr->startp[2], 2);
-	if (winRootPatternPtr->startp[6] != NULL) {
+    Tcl_RegExpRange(re, 0, &dummy, &tail);
+    Tcl_RegExpRange(re, 2, &drive, &dummy);
+    Tcl_RegExpRange(re, 3, &hostStart, &hostEnd);
+    Tcl_RegExpRange(re, 4, &shareStart, &shareEnd);
+    Tcl_RegExpRange(re, 6, &lastSlash, &dummy);
+
+    if (drive != NULL) {
+	Tcl_DStringAppend(resultPtr, drive, 2);
+	if (lastSlash != NULL) {
 	    Tcl_DStringAppend(resultPtr, "/", 1);
 	}
-    } else if (winRootPatternPtr->startp[4] != NULL) {
+    } else if (shareStart != NULL) {
 	Tcl_DStringAppend(resultPtr, "//", 2);
-	length = winRootPatternPtr->endp[3]
-	    - winRootPatternPtr->startp[3];
-	Tcl_DStringAppend(resultPtr, winRootPatternPtr->startp[3], length);
+	length = hostEnd - hostStart;
+	Tcl_DStringAppend(resultPtr, hostStart, length);
 	Tcl_DStringAppend(resultPtr, "/", 1);
-	length = winRootPatternPtr->endp[4]
-	    - winRootPatternPtr->startp[4];
-	Tcl_DStringAppend(resultPtr, winRootPatternPtr->startp[4], length);
+	length = shareEnd - shareStart;
+	Tcl_DStringAppend(resultPtr, shareStart, length);
     } else {
 	Tcl_DStringAppend(resultPtr, "/", 1);
     }
-    return winRootPatternPtr->endp[0];
+    return tail;
 }
 
 /*
@@ -197,7 +234,9 @@ Tcl_PathType
 Tcl_GetPathType(path)
     CONST char *path;
 {
+    ThreadSpecificData *tsdPtr;
     Tcl_PathType type = TCL_PATH_ABSOLUTE;
+    Tcl_RegExp re;
 
     switch (tclPlatform) {
    	case TCL_PLATFORM_UNIX:
@@ -214,45 +253,51 @@ Tcl_GetPathType(path)
 	    if (path[0] == ':') {
 		type = TCL_PATH_RELATIVE;
 	    } else if (path[0] != '~') {
+		tsdPtr = TCL_TSD_INIT(&dataKey);
 
 		/*
 		 * Since we have eliminated the easy cases, use the
 		 * root pattern to look for the other types.
 		 */
 
-		if (!macRootPatternPtr) {
-		    macRootPatternPtr = TclRegComp(MAC_ROOT_PATTERN);
-		    if (!initialized) {
-			Tcl_CreateExitHandler(FileNameCleanup, NULL);
-			initialized = 1;
-		    }
-		}
-		if (!TclRegExec(macRootPatternPtr, (char *) path, (char *) path)
-			|| (macRootPatternPtr->startp[2] != NULL)) {
+		FileNameInit();
+		re = Tcl_GetRegExpFromObj(NULL, tsdPtr->macRootPatternPtr,
+			REG_ADVANCED);
+
+		if (!Tcl_RegExpExec(NULL, re, path, path)) {
 		    type = TCL_PATH_RELATIVE;
+		} else {
+		    char *unixRoot, *dummy;
+
+		    Tcl_RegExpRange(re, 2, &unixRoot, &dummy);
+		    if (unixRoot) {
+			type = TCL_PATH_RELATIVE;
+		    }
 		}
 	    }
 	    break;
 	
 	case TCL_PLATFORM_WINDOWS:
 	    if (path[0] != '~') {
+		tsdPtr = TCL_TSD_INIT(&dataKey);
 
 		/*
 		 * Since we have eliminated the easy cases, check for
 		 * drive relative paths using the regular expression.
 		 */
 
-		if (!winRootPatternPtr) {
-		    winRootPatternPtr = TclRegComp(WIN_ROOT_PATTERN);
-		    if (!initialized) {
-			Tcl_CreateExitHandler(FileNameCleanup, NULL);
-			initialized = 1;
-		    }
-		}
-		if (TclRegExec(winRootPatternPtr, (char *) path, (char *) path)) {
-		    if (winRootPatternPtr->startp[5]
-			    || (winRootPatternPtr->startp[2]
-				    && !(winRootPatternPtr->startp[6]))) {
+		FileNameInit();
+		re = Tcl_GetRegExpFromObj(NULL, tsdPtr->winRootPatternPtr,
+			REG_ADVANCED);
+
+		if (Tcl_RegExpExec(NULL, re, path, path)) {
+		    char *drive, *dummy, *unixRoot, *lastSlash;
+
+		    Tcl_RegExpRange(re, 2, &drive, &dummy);
+		    Tcl_RegExpRange(re, 5, &unixRoot, &dummy);
+		    Tcl_RegExpRange(re, 6, &lastSlash, &dummy);
+		    
+		    if (unixRoot || (drive && !lastSlash)) {
 			type = TCL_PATH_VOLUME_RELATIVE;
 		    }
 		} else {
@@ -295,12 +340,13 @@ Tcl_SplitPath(path, argcPtr, argvPtr)
     CONST char *path;		/* Pointer to string containing a path. */
     int *argcPtr;		/* Pointer to location to fill in with
 				 * the number of elements in the path. */
-    CONST char ***argvPtr;	/* Pointer to place to store pointer to array
+    char ***argvPtr;		/* Pointer to place to store pointer to array
 				 * of pointers to path elements. */
 {
     int i, size;
     char *p;
     Tcl_DString buffer;
+
     Tcl_DStringInit(&buffer);
 
     /*
@@ -341,7 +387,7 @@ Tcl_SplitPath(path, argcPtr, argvPtr)
      * DString plus the argv pointers and the terminating NULL pointer.
      */
 
-    *argvPtr = (CONST char **) ckalloc((unsigned)
+    *argvPtr = (char **) ckalloc((unsigned)
 	    ((((*argcPtr) + 1) * sizeof(char *)) + size));
 
     /*
@@ -511,82 +557,92 @@ SplitMacPath(path, bufPtr)
     int isMac = 0;		/* 1 if is Mac-style, 0 if Unix-style path. */
     int i, length;
     CONST char *p, *elementStart;
+    Tcl_RegExp re;
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
     /*
      * Initialize the path name parser for Macintosh path names.
      */
 
-    if (macRootPatternPtr == NULL) {
-	macRootPatternPtr = TclRegComp(MAC_ROOT_PATTERN);
-	if (!initialized) {
-	    Tcl_CreateExitHandler(FileNameCleanup, NULL);
-	    initialized = 1;
-	}
-    }
+    FileNameInit();
 
     /*
      * Match the root portion of a Mac path name.
      */
 
     i = 0;			/* Needed only to prevent gcc warnings. */
-    if (TclRegExec(macRootPatternPtr, (char *) path, (char *) path) == 1) {
+
+    re = Tcl_GetRegExpFromObj(NULL, tsdPtr->macRootPatternPtr, REG_ADVANCED);
+
+    if (Tcl_RegExpExec(NULL, re, path, path) == 1) {
+	char *start, *end;
+
 	/*
 	 * Treat degenerate absolute paths like / and /../.. as
 	 * Mac relative file names for lack of anything else to do.
 	 */
 
-	if (macRootPatternPtr->startp[2] != NULL) {
+	Tcl_RegExpRange(re, 2, &start, &end);
+	if (start) {
 	    Tcl_DStringAppend(bufPtr, ":", 1);
-	    Tcl_DStringAppend(bufPtr, path, macRootPatternPtr->endp[0]
-		    - macRootPatternPtr->startp[0] + 1);
+	    Tcl_RegExpRange(re, 0, &start, &end);
+	    Tcl_DStringAppend(bufPtr, path, end - start + 1);
 	    return Tcl_DStringValue(bufPtr);
 	}
 
-	if (macRootPatternPtr->startp[5] != NULL) {
-
+	Tcl_RegExpRange(re, 5, &start, &end);
+	if (start) {
 	    /*
 	     * Unix-style tilde prefixed paths.
 	     */
 
 	    isMac = 0;
 	    i = 5;
-	} else if (macRootPatternPtr->startp[7] != NULL) {
+	} else {
+	    Tcl_RegExpRange(re, 7, &start, &end);
+	    if (start) {
+		/*
+		 * Mac-style tilde prefixed paths.
+		 */
 
-	    /*
-	     * Mac-style tilde prefixed paths.
-	     */
+		isMac = 1;
+		i = 7;
+	    } else {
+		Tcl_RegExpRange(re, 10, &start, &end);
+		if (start) {
 
-	    isMac = 1;
-	    i = 7;
-	} else if (macRootPatternPtr->startp[10] != NULL) {
+		    /*
+		     * Normal Unix style paths.
+		     */
 
-	    /*
-	     * Normal Unix style paths.
-	     */
+		    isMac = 0;
+		    i = 10;
+		} else {
+		    Tcl_RegExpRange(re, 12, &start, &end);
+		    if (start) {
 
-	    isMac = 0;
-	    i = 10;
-	} else if (macRootPatternPtr->startp[12] != NULL) {
+			/*
+			 * Normal Mac style paths.
+			 */
 
-	    /*
-	     * Normal Mac style paths.
-	     */
-
-	    isMac = 1;
-	    i = 12;
+			isMac = 1;
+			i = 12;
+		    }
+		}
+	    }
 	}
 
-	length = macRootPatternPtr->endp[i]
-	    - macRootPatternPtr->startp[i];
+	Tcl_RegExpRange(re, i, &start, &end);
+	length = end - start;
 
 	/*
 	 * Append the element and terminate it with a : and a null.  Note that
 	 * we are forcing the DString to contain an extra null at the end.
 	 */
 
-	Tcl_DStringAppend(bufPtr, macRootPatternPtr->startp[i], length);
+	Tcl_DStringAppend(bufPtr, start, length);
 	Tcl_DStringAppend(bufPtr, ":", 2);
-	p = macRootPatternPtr->endp[i];
+	p = end;
     } else {
 	isMac = (strchr(path, ':') != NULL);
 	p = path;
@@ -685,13 +741,13 @@ SplitMacPath(path, bufPtr)
 char *
 Tcl_JoinPath(argc, argv, resultPtr)
     int argc;
-    CONST char * CONST argv[];
+    char **argv;
     Tcl_DString *resultPtr;	/* Pointer to previously initialized DString. */
 {
     int oldLength, length, i, needsSep;
     Tcl_DString buffer;
-    CONST char *p;
     char c, *dest;
+    CONST char *p;
 
     Tcl_DStringInit(&buffer);
     oldLength = Tcl_DStringLength(resultPtr);
@@ -885,25 +941,27 @@ Tcl_JoinPath(argc, argv, resultPtr)
 }
 
 /*
- *----------------------------------------------------------------------
+ *---------------------------------------------------------------------------
  *
  * Tcl_TranslateFileName --
  *
  *	Converts a file name into a form usable by the native system
- *	interfaces.  If the name starts with a tilde, it will produce
- *	a name where the tilde and following characters have been
- *	replaced by the home directory location for the named user.
+ *	interfaces.  If the name starts with a tilde, it will produce a
+ *	name where the tilde and following characters have been replaced
+ *	by the home directory location for the named user.
  *
  * Results:
- *	The result is a pointer to a static string containing
- *	the new name.  If there was an error in processing the
- *	name, then an error message is left in interp->result
- *	and the return value is NULL.  The result will be stored
- *	in bufferPtr; the caller must call Tcl_DStringFree(bufferPtr)
- *	to free the name if the return value was not NULL.
+ *	The return value is a pointer to a string containing the name
+ *	after tilde substitution.  If there was no tilde substitution,
+ *	the return value is a pointer to a copy of the original string.
+ *	If there was an error in processing the name, then an error
+ *	message is left in the interp's result (if interp was not NULL)
+ *	and the return value is NULL.  Space for the return value is
+ *	allocated in bufferPtr; the caller must call Tcl_DStringFree()
+ *	to free the space if the return value was not NULL.
  *
  * Side effects:
- *	Information may be left in bufferPtr.
+ *	None.
  *
  *----------------------------------------------------------------------
  */
@@ -912,13 +970,12 @@ char *
 Tcl_TranslateFileName(interp, name, bufferPtr)
     Tcl_Interp *interp;		/* Interpreter in which to store error
 				 * message (if necessary). */
-    CONST char *name;		/* File name, which may begin with "~"
-				 * (to indicate current user's home directory)
-				 * or "~<user>" (to indicate any user's
-				 * home directory). */
-    Tcl_DString *bufferPtr;	/* May be used to hold result.  Must not hold
-				 * anything at the time of the call, and need
-				 * not even be initialized. */
+    CONST char *name;		/* File name, which may begin with "~" (to
+				 * indicate current user's home directory) or
+				 * "~<user>" (to indicate any user's home
+				 * directory). */
+    Tcl_DString *bufferPtr;	/* Uninitialized or free DString filled
+				 * with name after tilde substitution. */
 {
     register char *p;
 
@@ -931,11 +988,11 @@ Tcl_TranslateFileName(interp, name, bufferPtr)
 	char **argv;
 	Tcl_DString temp;
 
-	Tcl_SplitPath(name, &argc, (CONST char ***) &argv);
+	Tcl_SplitPath(name, &argc, (char ***) &argv);
 	
 	/*
-	 * Strip the trailing ':' off of a Mac path
-	 * before passing the user name to DoTildeSubst.
+	 * Strip the trailing ':' off of a Mac path before passing the user
+	 * name to DoTildeSubst.
 	 */
 
 	if (tclPlatform == TCL_PLATFORM_MAC) {
@@ -951,12 +1008,12 @@ Tcl_TranslateFileName(interp, name, bufferPtr)
 	    return NULL;
 	}
 	Tcl_DStringInit(bufferPtr);
-	Tcl_JoinPath(argc, (CONST char **) argv, bufferPtr);
+	Tcl_JoinPath(argc, (char **) argv, bufferPtr);
 	Tcl_DStringFree(&temp);
 	ckfree((char*)argv);
     } else {
 	Tcl_DStringInit(bufferPtr);
-	Tcl_JoinPath(1, &name, bufferPtr);
+	Tcl_JoinPath(1, (char **) &name, bufferPtr);
     }
 
     /*
@@ -1052,9 +1109,10 @@ TclGetExtension(name)
  * Results:
  *	The result is a pointer to a static string containing the home
  *	directory in native format.  If there was an error in processing
- *	the substitution, then an error message is left in interp->result
- *	and the return value is NULL.  On success, the results are appended
- * 	to resultPtr, and the contents of resultPtr are returned.
+ *	the substitution, then an error message is left in the interp's
+ *	result and the return value is NULL.  On success, the results
+ *	are appended to resultPtr, and the contents of resultPtr are
+ *	returned.
  *
  * Side effects:
  *	Information may be left in resultPtr.
@@ -1068,14 +1126,15 @@ DoTildeSubst(interp, user, resultPtr)
 				 * message (if necessary). */
     CONST char *user;		/* Name of user whose home directory should be
 				 * substituted, or "" for current user. */
-    Tcl_DString *resultPtr;	/* May be used to hold result.  Must not hold
-				 * anything at the time of the call, and need
-				 * not even be initialized. */
+    Tcl_DString *resultPtr;	/* Initialized DString filled with name
+				 * after tilde substitution. */
 {
-    CONST char *dir;
+    char *dir;
 
     if (*user == '\0') {
-	dir = TclGetEnv("HOME");
+	Tcl_DString dirString;
+	
+	dir = TclGetEnv("HOME", &dirString);
 	if (dir == NULL) {
 	    if (interp) {
 		Tcl_ResetResult(interp);
@@ -1085,10 +1144,9 @@ DoTildeSubst(interp, user, resultPtr)
 	    return NULL;
 	}
 	Tcl_JoinPath(1, &dir, resultPtr);
+	Tcl_DStringFree(&dirString);
     } else {
-	
-	/* lint, TclGetuserHome() always NULL under windows. */
-	if (TclGetUserHome(user, resultPtr) == NULL) {	
+	if (TclpGetUserHome(user, resultPtr) == NULL) {	
 	    if (interp) {
 		Tcl_ResetResult(interp);
 		Tcl_AppendResult(interp, "user \"", user, "\" doesn't exist",
@@ -1103,7 +1161,7 @@ DoTildeSubst(interp, user, resultPtr)
 /*
  *----------------------------------------------------------------------
  *
- * Tcl_GlobCmd --
+ * Tcl_GlobObjCmd --
  *
  *	This procedure is invoked to process the "glob" Tcl command.
  *	See the user documentation for details on what it does.
@@ -1119,42 +1177,112 @@ DoTildeSubst(interp, user, resultPtr)
 
 	/* ARGSUSED */
 int
-Tcl_GlobCmd(dummy, interp, argc, argv)
+Tcl_GlobObjCmd(dummy, interp, objc, objv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;			/* Current interpreter. */
-    int argc;				/* Number of arguments. */
-    char **argv;			/* Argument strings. */
+    int objc;				/* Number of arguments. */
+    Tcl_Obj *CONST objv[];		/* Argument objects. */
 {
-    int i, noComplain, firstArg;
-    char c;
-    int result = TCL_OK;
-    Tcl_DString buffer;
-    char *separators, *head, *tail;
-
-    noComplain = 0;
-    for (firstArg = 1; (firstArg < argc) && (argv[firstArg][0] == '-');
-	    firstArg++) {
-	if (strcmp(argv[firstArg], "-nocomplain") == 0) {
-	    noComplain = 1;
-	} else if (strcmp(argv[firstArg], "--") == 0) {
-	    firstArg++;
-	    break;
-	} else {
-	    Tcl_AppendResult(interp, "bad switch \"", argv[firstArg],
-		    "\": must be -nocomplain or --", (char *) NULL);
-	    return TCL_ERROR;
+    int index, i, globFlags, pathlength, length, join, dir, result;
+    char *string, *pathOrDir, *separators;
+    Tcl_Obj *typePtr, *resultPtr, *look;
+    Tcl_DString prefix, directory;
+    static char *options[] = {"-dir", "-join", "-nocomplain", 
+      "-path", "-types", "--", NULL};
+    enum options {GLOB_DIR, GLOB_JOIN, GLOB_NOCOMPLAIN, 
+      GLOB_PATH, GLOB_TYPE, GLOB_LAST};
+    enum pathDirOptions {PATH_NONE = -1 , PATH_GENERAL = 0, PATH_DIR = 1};
+    GlobTypeData *globTypes = NULL;
+    
+    globFlags = 0;
+    join = 0;
+    dir = PATH_NONE;
+    pathOrDir = NULL;
+    typePtr = NULL;
+    resultPtr = Tcl_GetObjResult(interp);
+    for (i = 1; i < objc; i++) {
+	if (Tcl_GetIndexFromObj(interp, objv[i], options, "option", 0, &index)
+		!= TCL_OK) {
+	    string = Tcl_GetStringFromObj(objv[i],&length);
+	    if(string[0] == '-') {
+		/* It looks like the command contains an option so signal an error */
+		return TCL_ERROR;
+	    } else {
+		/* This clearly isn't an option; assume it's the first glob pattern */
+		/* We must clear the error */
+		Tcl_ResetResult(interp);
+		break;
+	    }
+	}
+	switch (index) {
+	    case GLOB_NOCOMPLAIN:			/* -nocomplain */
+	        globFlags |= GLOBMODE_NO_COMPLAIN;
+		break;
+	    case GLOB_DIR:			/* -dir */
+		if (i == (objc-1)) {
+		    Tcl_AppendToObj(resultPtr,
+			    "\"-dir\" option must be followed by a directory",
+			    -1);
+		    return TCL_ERROR;
+		}
+		if(dir != -1) {
+		    Tcl_AppendToObj(resultPtr,
+			    "\"-dir\" option is not compatible with \"-path\"",
+			    -1);
+		    return TCL_ERROR;
+		}
+		dir = PATH_DIR;
+		globFlags |= GLOBMODE_DIR;
+		pathOrDir = Tcl_GetStringFromObj(objv[i+1],&pathlength);
+		i++;
+		break;
+	    case GLOB_PATH:			/* -path */
+	        if (i == (objc-1)) {
+		    Tcl_AppendToObj(resultPtr,
+			    "\"-path\" option must be followed by a path",
+			    -1);
+		    return TCL_ERROR;
+		}
+		if(dir != -1) {
+		    Tcl_AppendToObj(resultPtr,
+			    "\"-path\" option is not compatible with \"-dir\"",
+			    -1);
+		    return TCL_ERROR;
+		}
+		dir = PATH_GENERAL;
+		pathOrDir = Tcl_GetStringFromObj(objv[i+1],&pathlength);
+		i++;
+		break;
+	    case GLOB_JOIN:			/* -join */
+		join = 1;
+		break;
+	    case GLOB_TYPE:			/* -type */
+	        if (i == (objc-1)) {
+		    Tcl_AppendToObj(resultPtr,
+			    "\"-types\" option must be followed by a list of types",
+			    -1);
+		    return TCL_ERROR;
+		}
+		typePtr = objv[i+1];
+		if (Tcl_ListObjLength(interp,typePtr,&length) != TCL_OK) {
+		    return TCL_ERROR;
+		}
+		i++;
+		break;
+	    case GLOB_LAST:			/* -- */
+	        i++;
+		goto endOfForLoop;
+		break;
 	}
     }
-    if (firstArg >= argc) {
-	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		" ?switches? name ?name ...?\"", (char *) NULL);
+    endOfForLoop:
+    if (objc - i < 1) {
+        Tcl_WrongNumArgs(interp, 1, objv, "?switches? name ?name ...?");
 	return TCL_ERROR;
     }
 
-    Tcl_DStringInit(&buffer);
-    separators = NULL;		/* Needed only to prevent gcc warnings. */
-    for (i = firstArg; i < argc; i++) {
-	switch (tclPlatform) {
+    separators = NULL;		/* lint. */
+    switch (tclPlatform) {
 	case TCL_PLATFORM_UNIX:
 	    separators = "/";
 	    break;
@@ -1162,102 +1290,326 @@ Tcl_GlobCmd(dummy, interp, argc, argv)
 	    separators = "/\\:";
 	    break;
 	case TCL_PLATFORM_MAC:
-	    separators = (strchr(argv[i], ':') == NULL) ? "/" : ":";
+	    separators = (pathOrDir != NULL && (strchr(pathOrDir, ':') == NULL))
+		? "/" : ":";
 	    break;
-	}
-
-	Tcl_DStringSetLength(&buffer, 0);
+    }
+    if (dir == PATH_GENERAL) {
+	char *last;
 
 	/*
-	 * Perform tilde substitution, if needed.
+	 * Find the last path separator in the path
 	 */
+	last = pathOrDir + pathlength;
+	for (; last != pathOrDir; last--) {
+	    if (strchr(separators, *(last-1)) != NULL) {
+		break;
+	    }
+	}
+	if(last == pathOrDir + pathlength) {
+	    /* It's really a directory */
+	    dir = 1;
+	} else {
+	    Tcl_DString pref;
+	    char *search, *find;
+	    Tcl_DStringInit(&pref);
+	    Tcl_DStringInit(&directory);
+	    if(last == pathOrDir) {
+		/* The whole thing is a prefix */
+		Tcl_DStringAppend(&pref,pathOrDir,-1);
+		pathOrDir = NULL;
+	    } else {
+		/* Have to split off the end */
+		Tcl_DStringAppend(&pref,last,pathOrDir+pathlength-last);
+		Tcl_DStringAppend(&directory,pathOrDir,last-pathOrDir-1);
+		pathOrDir = Tcl_DStringValue(&directory);
+	    }
+	    /* Need to quote 'prefix' */
+	    Tcl_DStringInit(&prefix);
+	    search = Tcl_DStringValue(&pref);
+	    while((find = (strpbrk(search, "\\[]*?{}"))) != NULL) {
+	        Tcl_DStringAppend(&prefix,search,find-search);
+	        Tcl_DStringAppend(&prefix,"\\",1);
+	        Tcl_DStringAppend(&prefix,find,1);
+	        search = find+1;
+	        if(*search == '\0') {
+	            break;
+	        }
+	    }
+	    if(*search != '\0') {
+		Tcl_DStringAppend(&prefix,search,-1);
+	    }
+	    Tcl_DStringFree(&pref);
+	}
+    }
 
-	if (argv[i][0] == '~') {
-	    char *p;
+    if(typePtr != NULL) {
+	/* 
+	 * The rest of the possible type arguments 
+	 * are currently ignored.
+	 */
+	while(--length >= 0) {
+	    Tcl_ListObjIndex(interp,typePtr,length,&look);
+	    if(!strcmp("d",Tcl_GetString(look))) {
+		globFlags |= GLOBMODE_DIR_ONLY;
+	    }
+	}
+    }
 
-	    /*
-	     * Find the first path separator after the tilde.
-	     */
+    /* 
+     * Now we perform the actual glob below.  This may involve joining
+     * together the pattern arguments, dealing with particular file types
+     * etc.  We use a 'goto' to ensure we free any memory allocated along
+     * the way.
+     */
+    objc -= i;
+    objv += i;
+    /* 
+     * We re-retrieve this, in case it was changed in 
+     * the Tcl_ResetResult above 
+     */
+    resultPtr = Tcl_GetObjResult(interp);
+    result = TCL_OK;
+    if(join) {
+	if(dir != PATH_GENERAL) {
+	    Tcl_DStringInit(&prefix);
+	}
+	for (i = 0; i < objc; i++) {
+	    string = Tcl_GetStringFromObj(objv[i],&length);
+	    Tcl_DStringAppend(&prefix,string,length);
+	    if (i != objc -1) {
+		Tcl_DStringAppend(&prefix,separators,1);
+	    }
+	}
+	if(globFlags & GLOBMODE_DIR_ONLY) {
+	    Tcl_DStringAppend(&prefix,separators,1);
+	}
+	if (TclGlob(interp, Tcl_DStringValue(&prefix), pathOrDir, globFlags, globTypes) != TCL_OK) {
+	    result = TCL_ERROR;
+	    goto endOfGlob;
+	}
+    } else {
+	if ((dir == PATH_GENERAL) || (globFlags & GLOBMODE_DIR_ONLY)) {
+	    Tcl_DString str;
+	    for (i = 0; i < objc; i++) {
+		Tcl_DStringInit(&str);
+		if(dir == PATH_GENERAL) {
+		    Tcl_DStringAppend(&str,Tcl_DStringValue(&prefix),Tcl_DStringLength(&prefix));
+		}
+		string = Tcl_GetStringFromObj(objv[i],&length);
+		Tcl_DStringAppend(&str,string,length);
+		if(globFlags & GLOBMODE_DIR_ONLY) {
+		    Tcl_DStringAppend(&str,separators,1);
+		}
+		if (TclGlob(interp, Tcl_DStringValue(&str), pathOrDir, globFlags, globTypes) != TCL_OK) {
+		    result = TCL_ERROR;
+		    Tcl_DStringFree(&str);
+		    goto endOfGlob;
+		}
+	    }
+	    Tcl_DStringFree(&str);
+	} else {
+	    for (i = 0; i < objc; i++) {
+		string = Tcl_GetString(objv[i]);
+		if (TclGlob(interp, string, pathOrDir, globFlags, globTypes) != TCL_OK) {
+		    result = TCL_ERROR;
+		    goto endOfGlob;
+		}
+	    }
+	}
+    }
+    if ((globFlags & GLOBMODE_NO_COMPLAIN) == 0) {
+	if(Tcl_ListObjLength(interp, Tcl_GetObjResult(interp), &length) != TCL_OK) {
+	    /* This should never happen.  Maybe we should be more dramatic */
+	    result = TCL_ERROR;
+	    goto endOfGlob;
+	}
+	if (length == 0) {
+	    Tcl_AppendResult(interp, "no files matched glob pattern",
+		    (join || (objc == 1)) ? " \"" : "s \"", (char *) NULL);
+	    if(join) {
+		Tcl_AppendResult(interp,Tcl_DStringValue(&prefix), (char *) NULL);
+	    } else {
+		char *sep = "";
+		for (i = 0; i < objc; i++) {
+		    string = Tcl_GetString(objv[i]);
+		    Tcl_AppendResult(interp, sep, string, (char *) NULL);
+		    sep = " ";
+		}
+	    }
+	    Tcl_AppendResult(interp, "\"", (char *) NULL);
+	    result = TCL_ERROR;
+	}
+    }
+  endOfGlob:
+    if(join || (dir == PATH_GENERAL)) {
+	Tcl_DStringFree(&prefix);
+	if(dir == PATH_GENERAL) {
+	    Tcl_DStringFree(&directory);
+	}
+    }
+    return result;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclGlob --
+ *
+ *	This procedure prepares arguments for the TclDoGlob call.
+ *	It sets the separator string based on the platform, performs
+ *      tilde substitution, and calls TclDoGlob.
+ *
+ * Results:
+ *	The return value is a standard Tcl result indicating whether
+ *	an error occurred in globbing.  After a normal return the
+ *	result in interp (set by TclDoGlob) holds all of the file names
+ *	given by the dir and rem arguments.  After an error the
+ *	result in interp will hold an error message.
+ *
+ * Side effects:
+ *	The currentArgString is written to.
+ *
+ *----------------------------------------------------------------------
+ */
 
-	    for (tail = argv[i]; *tail != '\0'; tail++) {
-		if (*tail == '\\') {
-		    if (strchr(separators, tail[1]) != NULL) {
-			break;
-		    }
-		} else if (strchr(separators, *tail) != NULL) {
+	/* ARGSUSED */
+int
+TclGlob(interp, pattern, unquotedPrefix, globFlags, types)
+    Tcl_Interp *interp;		/* Interpreter for returning error message
+				 * or appending list of matching file names. */
+    char *pattern;		/* Glob pattern to match. Must not refer
+				 * to a static string. */
+    char *unquotedPrefix;	/* Prefix to glob pattern, if non-null, which
+                             	 * is considered literally.  May be static. */
+    int globFlags;		/* Stores or'ed combination of flags */
+    GlobTypeData *types;	/* Struct containing acceptable types.
+				 * May be NULL. */
+{
+    char *separators;
+    char *head, *tail, *start;
+    char c;
+    int result;
+    Tcl_DString buffer;
+
+    separators = NULL;		/* lint. */
+    switch (tclPlatform) {
+	case TCL_PLATFORM_UNIX:
+	    separators = "/";
+	    break;
+	case TCL_PLATFORM_WINDOWS:
+	    separators = "/\\:";
+	    break;
+	case TCL_PLATFORM_MAC:
+	    separators = (strchr(pattern, ':') == NULL)
+		? "/" : ":";
+	    break;
+    }
+
+    Tcl_DStringInit(&buffer);
+    if(unquotedPrefix != NULL) {
+	start = unquotedPrefix;
+    } else {
+	start = pattern;
+    }
+
+    /*
+     * Perform tilde substitution, if needed.
+     */
+
+    if (start[0] == '~') {
+	
+	/*
+	 * Find the first path separator after the tilde.
+	 */
+	for (tail = start; *tail != '\0'; tail++) {
+	    if (*tail == '\\') {
+		if (strchr(separators, tail[1]) != NULL) {
 		    break;
 		}
+	    } else if (strchr(separators, *tail) != NULL) {
+		break;
 	    }
-
-	    /*
-	     * Determine the home directory for the specified user.  Note that
-	     * we don't allow special characters in the user name.
-	     */
-
-	    c = *tail;
-	    *tail = '\0';
-	    p = strpbrk(argv[i]+1, "\\[]*?{}");
-	    if (p == NULL) {
-		head = DoTildeSubst(interp, argv[i]+1, &buffer);
-	    } else {
-		if (!noComplain) {
-		    Tcl_ResetResult(interp);
-		    Tcl_AppendResult(interp, "globbing characters not ",
-			    "supported in user names", (char *) NULL);
-		}
-		head = NULL;
-	    }
-	    *tail = c;
-	    if (head == NULL) {
-		if (noComplain) {
-		    Tcl_ResetResult(interp);
-		    continue;
-		} else {
-		    result = TCL_ERROR;
-		    goto done;
-		}
-	    }
-	    if (head != Tcl_DStringValue(&buffer)) {
-		Tcl_DStringAppend(&buffer, head, -1);
-	    }
-	} else {
-	    tail = argv[i];
 	}
 
-	result = TclDoGlob(interp, separators, &buffer, tail);
-	if (result != TCL_OK) {
-	    if (noComplain) {
+	/*
+	 * Determine the home directory for the specified user.  Note that
+	 * we don't allow special characters in the user name.
+	 */
+	
+	c = *tail;
+	*tail = '\0';
+	/*
+	 * I don't think we need to worry about special characters in
+	 * the user name anymore (Vince Darley, June 1999), since the
+	 * new code is designed to handle special chars.
+	 */
+	#ifndef NOT_NEEDED_ANYMORE
+	head = DoTildeSubst(interp, start+1, &buffer);
+	#else
+	
+	if (strpbrk(start+1, "\\[]*?{}") == NULL) {
+	    head = DoTildeSubst(interp, start+1, &buffer);
+	} else {
+	    if (!(globFlags & GLOBMODE_NO_COMPLAIN)) {
+		Tcl_ResetResult(interp);
+		Tcl_AppendResult(interp, "globbing characters not ",
+			"supported in user names", (char *) NULL);
+	    }
+	    head = NULL;
+	}
+	#endif
+	*tail = c;
+	if (head == NULL) {
+	    if (globFlags & GLOBMODE_NO_COMPLAIN) {
 		/*
 		 * We should in fact pass down the nocomplain flag 
-		 * or save the interp result or use another mecanism
+		 * or save the interp result or use another mechanism
 		 * so the interp result is not mangled on errors in that case.
 		 * but that would a bigger change than reasonable for a patch
 		 * release.
 		 * (see fileName.test 15.2-15.4 for expected behaviour)
 		 */
 		Tcl_ResetResult(interp);
-		result = TCL_OK;
-		continue;
+		return TCL_OK;
 	    } else {
-		goto done;
+		return TCL_ERROR;
+	    }
+	}
+	if (head != Tcl_DStringValue(&buffer)) {
+	    Tcl_DStringAppend(&buffer, head, -1);
+	}
+	if(unquotedPrefix != NULL) {
+	    Tcl_DStringAppend(&buffer, tail, -1);
+	    tail = pattern;
+	}
+    } else {
+	tail = pattern;
+	if(unquotedPrefix != NULL) {
+	    Tcl_DStringAppend(&buffer,unquotedPrefix,-1);
+	}
+    }
+    /* 
+     * If the prefix is a directory, make sure it ends in a directory
+     * separator.
+     */
+    if(unquotedPrefix != NULL) {
+	if (globFlags & GLOBMODE_DIR) {
+	    c = Tcl_DStringValue(&buffer)[Tcl_DStringLength(&buffer)-1];
+	    if (strchr(separators, c) == NULL) {
+		Tcl_DStringAppend(&buffer,separators,1);
 	    }
 	}
     }
 
-    if ((*interp->result == 0) && !noComplain) {
-	char *sep = "";
-
-	Tcl_AppendResult(interp, "no files matched glob pattern",
-		(argc == 2) ? " \"" : "s \"", (char *) NULL);
-	for (i = firstArg; i < argc; i++) {
-	    Tcl_AppendResult(interp, sep, argv[i], (char *) NULL);
-	    sep = " ";
-	}
-	Tcl_AppendResult(interp, "\"", (char *) NULL);
-	result = TCL_ERROR;
-    }
-done:
+    result = TclDoGlob(interp, separators, &buffer, tail, types);
     Tcl_DStringFree(&buffer);
+    if (result != TCL_OK) {
+	if (globFlags & GLOBMODE_NO_COMPLAIN) {
+	    Tcl_ResetResult(interp);
+	    return TCL_OK;
+	}
+    }
     return result;
 }
 
@@ -1321,7 +1673,11 @@ SkipToChar(stringPtr, match)
  *	This recursive procedure forms the heart of the globbing
  *	code.  It performs a depth-first traversal of the tree
  *	given by the path name to be globbed.  The directory and
- *	remainder are assumed to be native format paths.
+ *	remainder are assumed to be native format paths.  The prefix 
+ *	contained in 'headPtr' is not used as a glob pattern, simply
+ *	as a path specifier, so it can contain unquoted glob-sensitive
+ *	characters (if the directories to which it points contain
+ *	such strange characters).
  *
  * Results:
  *	The return value is a standard Tcl result indicating whether
@@ -1337,19 +1693,23 @@ SkipToChar(stringPtr, match)
  */
 
 int
-TclDoGlob(interp, separators, headPtr, tail)
+TclDoGlob(interp, separators, headPtr, tail, types)
     Tcl_Interp *interp;		/* Interpreter to use for error reporting
 				 * (e.g. unmatched brace). */
     char *separators;		/* String containing separator characters
 				 * that should be used to identify globbing
 				 * boundaries. */
     Tcl_DString *headPtr;	/* Completely expanded prefix. */
-    char *tail;			/* The unexpanded remainder of the path. */
+    char *tail;			/* The unexpanded remainder of the path.
+				 * Must not be a pointer to a static string. */
+    GlobTypeData *types;	/* List object containing list of acceptable types.
+				 * May be NULL. */
 {
     int baseLength, quoted, count;
     int result = TCL_OK;
-    char *p, *openBrace, *closeBrace, *name, *firstSpecialChar, savedChar;
+    char *name, *p, *openBrace, *closeBrace, *firstSpecialChar, savedChar;
     char lastChar = 0;
+    
     int length = Tcl_DStringLength(headPtr);
 
     if (length > 0) {
@@ -1377,7 +1737,6 @@ TclDoGlob(interp, separators, headPtr, tail)
      * for multiple separators, since they are special in Mac-style
      * paths.
      */
-
     switch (tclPlatform) {
 	case TCL_PLATFORM_MAC:
 	    if (*separators == '/') {
@@ -1501,7 +1860,7 @@ TclDoGlob(interp, separators, headPtr, tail)
 	    Tcl_DStringAppend(&newName, element, p-element);
 	    Tcl_DStringAppend(&newName, closeBrace+1, -1);
 	    result = TclDoGlob(interp, separators,
-		    headPtr, Tcl_DStringValue(&newName));
+		    headPtr, Tcl_DStringValue(&newName), NULL);
 	    if (result != TCL_OK) {
 		break;
 	    }
@@ -1520,6 +1879,12 @@ TclDoGlob(interp, separators, headPtr, tail)
      */
 
     if (*p != '\0') {
+
+	/*
+	 * Note that we are modifying the string in place.  This won't work
+	 * if the string is a static.
+	 */
+
 	 savedChar = *p;
 	 *p = '\0';
 	 firstSpecialChar = strpbrk(tail, "*[]?\\");
@@ -1533,15 +1898,15 @@ TclDoGlob(interp, separators, headPtr, tail)
 	 * Look for matching files in the current directory.  The
 	 * implementation of this function is platform specific, but may
 	 * recursively call TclDoGlob.  For each file that matches, it will
-	 * add the match onto the interp->result, or call TclDoGlob if there
+	 * add the match onto the interp's result, or call TclDoGlob if there
 	 * are more characters to be processed.
 	 */
 
-	return TclMatchFiles(interp, separators, headPtr, tail, p);
+	return TclpMatchFilesTypes(interp, separators, headPtr, tail, p, NULL);
     }
     Tcl_DStringAppend(headPtr, tail, p-tail);
     if (*p != '\0') {
-	return TclDoGlob(interp, separators, headPtr, p);
+	return TclDoGlob(interp, separators, headPtr, p, NULL);
     }
 
     /*
@@ -1551,21 +1916,25 @@ TclDoGlob(interp, separators, headPtr, tail)
      */
 
     switch (tclPlatform) {
-	case TCL_PLATFORM_MAC:
+	case TCL_PLATFORM_MAC: {
 	    if (strchr(Tcl_DStringValue(headPtr), ':') == NULL) {
 		Tcl_DStringAppend(headPtr, ":", 1);
 	    }
 	    name = Tcl_DStringValue(headPtr);
-	    if (TclAccess(name, F_OK) == 0) {
+	    if (TclpAccess(name, F_OK) == 0) {
 		if ((name[1] != '\0') && (strchr(name+1, ':') == NULL)) {
-		    Tcl_AppendElement(interp, name+1);
+		    Tcl_ListObjAppendElement(interp, Tcl_GetObjResult(interp), 
+					     Tcl_NewStringObj(name + 1,-1));
 		} else {
-		    Tcl_AppendElement(interp, name);
+		    Tcl_ListObjAppendElement(interp, Tcl_GetObjResult(interp), 
+					     Tcl_NewStringObj(name,-1));
 		}
 	    }
 	    break;
+	}
 	case TCL_PLATFORM_WINDOWS: {
 	    int exists;
+	    
 	    /*
 	     * We need to convert slashes to backslashes before checking
 	     * for the existence of the file.  Once we are done, we need
@@ -1587,18 +1956,20 @@ TclDoGlob(interp, separators, headPtr, tail)
 		}
 	    }
 	    name = Tcl_DStringValue(headPtr);
-	    exists = (TclAccess(name, F_OK) == 0);
+	    exists = (TclpAccess(name, F_OK) == 0);
+
 	    for (p = name; *p != '\0'; p++) {
 		if (*p == '\\') {
 		    *p = '/';
 		}
 	    }
 	    if (exists) {
-		Tcl_AppendElement(interp, name);
+		Tcl_ListObjAppendElement(interp, Tcl_GetObjResult(interp), 
+					 Tcl_NewStringObj(name,-1));
 	    }
 	    break;
 	}
-	case TCL_PLATFORM_UNIX:
+	case TCL_PLATFORM_UNIX: {
 	    if (Tcl_DStringLength(headPtr) == 0) {
 		if ((*name == '\\' && name[1] == '/') || (*name == '/')) {
 		    Tcl_DStringAppend(headPtr, "/", 1);
@@ -1607,10 +1978,12 @@ TclDoGlob(interp, separators, headPtr, tail)
 		}
 	    }
 	    name = Tcl_DStringValue(headPtr);
-	    if (TclAccess(name, F_OK) == 0) {
-		Tcl_AppendElement(interp, name);
+	    if (TclpAccess(name, F_OK) == 0) {
+		Tcl_ListObjAppendElement(interp, Tcl_GetObjResult(interp), 
+					 Tcl_NewStringObj(name,-1));
 	    }
 	    break;
+	}
     }
 
     return TCL_OK;
