@@ -10,17 +10,11 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: pkgd.c,v 1.2 1998/09/14 18:40:18 stanton Exp $
+ * RCS: @(#) $Id: pkgd.c,v 1.3 1999/03/11 21:47:40 stanton Exp $
  */
 
 #include "tcl.h"
 
-#ifdef TCL_USE_STUBS
-Tcl_StubTable *tclStubTable;
-#else
-#define Tcl_Required(a,b,c) Tcl_PkgRequire(a,"Tcl",b,c)
-#endif
-
 /*
  * Prototypes for procedures defined later in this file:
  */
@@ -52,7 +46,7 @@ Pkgd_SubObjCmd(dummy, interp, objc, objv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;			/* Current interpreter. */
     int objc;				/* Number of arguments. */
-    Tcl_Obj * CONST objv[];		/* Argument strings. */
+    Tcl_Obj * CONST objv[];		/* Argument objects. */
 {
     int first, second;
 
@@ -71,7 +65,7 @@ Pkgd_SubObjCmd(dummy, interp, objc, objv)
 /*
  *----------------------------------------------------------------------
  *
- * Pkgd_UnsafeObjCmd --
+ * Pkgd_UnsafeCmd --
  *
  *	This procedure is invoked to process the "pkgd_unsafe" Tcl command.
  *	It just returns a constant string.
@@ -114,21 +108,18 @@ Pkgd_UnsafeObjCmd(dummy, interp, objc, objv)
  */
 
 int
-#ifdef _USING_PROTOTYPES_
-Pkgd_Init (
-    Tcl_Interp *interp)		/* Interpreter in which the package is
-				 * to be made available. */
-#else
 Pkgd_Init(interp)
     Tcl_Interp *interp;		/* Interpreter in which the package is
 				 * to be made available. */
-#endif
 {
-    if (Tcl_Required(interp, "8", 0) == NULL) {
+    int code;
+
+    if (Tcl_InitStubs(interp, "8.0", 0) == NULL) {
 	return TCL_ERROR;
     }
-    if (Tcl_PkgProvide(interp, "Pkgd", "7.3") != TCL_OK) {
-	return TCL_ERROR;
+    code = Tcl_PkgProvide(interp, "Pkgd", "7.3");
+    if (code != TCL_OK) {
+	return code;
     }
     Tcl_CreateObjCommand(interp, "pkgd_sub", Pkgd_SubObjCmd, (ClientData) 0,
 	    (Tcl_CmdDeleteProc *) NULL);
@@ -155,21 +146,18 @@ Pkgd_Init(interp)
  */
 
 int
-#ifdef _USING_PROTOTYPES_
-Pkgd_SafeInit (
-    Tcl_Interp *interp)		/* Interpreter in which the package is
-				 * to be made available. */
-#else
 Pkgd_SafeInit(interp)
     Tcl_Interp *interp;		/* Interpreter in which the package is
 				 * to be made available. */
-#endif
 {
-    if (Tcl_Required(interp, "8", 0) == NULL) {
+    int code;
+
+    if (Tcl_InitStubs(interp, "8.0", 0) == NULL) {
 	return TCL_ERROR;
     }
-    if (Tcl_PkgProvide(interp, "Pkgd", "7.3") != TCL_OK) {
-	return TCL_ERROR;
+    code = Tcl_PkgProvide(interp, "Pkgd", "7.3");
+    if (code != TCL_OK) {
+	return code;
     }
     Tcl_CreateObjCommand(interp, "pkgd_sub", Pkgd_SubObjCmd, (ClientData) 0,
 	    (Tcl_CmdDeleteProc *) NULL);
