@@ -39,7 +39,7 @@ Pkge_Init(interp)
 				 * to be made available. */
 {
     static char script[] = "if 44 {open non_existent}";
-    if (Tcl_InitStubs(interp, TCL_VERSION, 1) == NULL) {
+    if (Tcl_InitStubs(interp, "8.0", 0) == NULL) {
 	return TCL_ERROR;
     }
     return Tcl_Eval(interp, script);
