@@ -91,7 +91,7 @@ TclpFindExecutable(argv0)
 /*
  *----------------------------------------------------------------------
  *
- * TclpMatchFiles --
+ * TclpMatchFilesTypes --
  *
  *	This routine is used by the globbing code to search a
  *	directory for all files which match a given pattern.
@@ -109,14 +109,16 @@ TclpFindExecutable(argv0)
  *---------------------------------------------------------------------- */
 
 int
-TclpMatchFiles(interp, separators, dirPtr, pattern, tail)
-    Tcl_Interp *interp;		/* Interpreter to receive results. */
-    char *separators;		/* Directory separators to pass to TclDoGlob. */
-    Tcl_DString *dirPtr;	/* Contains path to directory to search. */
-    char *pattern;		/* Pattern to match against. */
-    char *tail;			/* Pointer to end of pattern.  Tail must
-				 * point to a location in pattern. Must not
-				 * point to a static string. */
+TclpMatchFilesTypes(
+    Tcl_Interp *interp,		/* Interpreter to receive results. */
+    char *separators,		/* Directory separators to pass to TclDoGlob. */
+    Tcl_DString *dirPtr,	/* Contains path to directory to search. */
+    char *pattern,		/* Pattern to match against. */
+    char *tail,			/* Pointer to end of pattern.  Tail must
+				 * point to a location in pattern and must
+				 * not be static.*/
+    GlobTypeData *types)	/* Object containing list of acceptable types.
+				 * May be NULL. */
 {
     char drivePat[] = "?:\\";
     const char *message;
@@ -130,6 +132,10 @@ TclpMatchFiles(interp, separators, dirPtr, pattern, tail)
     BOOL found;
     Tcl_DString ds;
     TCHAR *nativeName;
+    Tcl_Obj *resultPtr;
+
+    /* Until we actually handle types, we do this to avoid warnings */
+    types = NULL;     
 
     /*
      * Convert the path to normalized form since some interfaces only
@@ -272,6 +278,7 @@ TclpMatchFiles(interp, separators, dirPtr, pattern, tail)
      * Now iterate over all of the files in the directory.
      */
 
+    resultPtr = Tcl_GetObjResult(interp);
     for (found = 1; found != 0; 
 	    found = (*tclWinProcs->findNextFileProc)(handle, &data)) {
 	TCHAR *nativeMatchResult;
@@ -323,7 +330,8 @@ TclpMatchFiles(interp, separators, dirPtr, pattern, tail)
 	Tcl_DStringFree(&ds);
 
 	if (tail == NULL) {
-	    Tcl_AppendElement(interp, Tcl_DStringValue(dirPtr));
+            Tcl_ListObjAppendElement(interp, resultPtr, 
+				     Tcl_NewStringObj(Tcl_DStringValue(dirPtr),Tcl_DStringLength(dirPtr)));
 	} else {
 	    nativeName = Tcl_WinUtfToTChar(Tcl_DStringValue(dirPtr), 
 		    Tcl_DStringLength(dirPtr), &ds);
@@ -332,7 +340,7 @@ TclpMatchFiles(interp, separators, dirPtr, pattern, tail)
 
 	    if (attr & FILE_ATTRIBUTE_DIRECTORY) {
 		Tcl_DStringAppend(dirPtr, "/", 1);
-		result = TclDoGlob(interp, separators, dirPtr, tail);
+		result = TclDoGlob(interp, separators, dirPtr, tail, types);
 		if (result != TCL_OK) {
 		    break;
 		}
@@ -356,6 +364,25 @@ TclpMatchFiles(interp, separators, dirPtr, pattern, tail)
     return TCL_ERROR;
 }
 
+/* 
+ * TclpMatchFiles --
+ * 
+ * This function is now obsolete.  Call the above function 
+ * 'TclpMatchFilesTypes' instead.
+ */
+int
+TclpMatchFiles(
+    Tcl_Interp *interp,		/* Interpreter to receive results. */
+    char *separators,		/* Directory separators to pass to TclDoGlob. */
+    Tcl_DString *dirPtr,	/* Contains path to directory to search. */
+    char *pattern,		/* Pattern to match against. */
+    char *tail)			/* Pointer to end of pattern.  Tail must
+				 * point to a location in pattern and must
+				 * not be static.*/
+{
+    return TclpMatchFilesTypes(interp,separators,dirPtr,pattern,tail,NULL);
+}
+
 /*
  *----------------------------------------------------------------------
  *
